(ns org.opencase.ordiagram.reporting.reporter
  (:use org.opencase.clojure.emfquery
        org.opencase.clojure.modelapi
        org.opencase.reporting.report
        org.opencase.reporting.base-reporter)
  (:require org.opencase.ordiagram.reporting.report
            clojure.string)
  (:import [org.opencase.ordiagram.reporting.report ORDiagram StatefulParticipant ServiceParticipant Level Slot Activity ServiceActivity State]))

(defn report-transition [trans]
  (let [con (first (mselect trans :constraint))
        target-id (element-id (connection-target trans))]
    (if (nil? con)
      target-id
      (str target-id " if " (first (mselect con :condition))))))

(defn create-state-goto [state visited]
  (let [targets (map connection-target (transitions state :out))]
    (if (and (nil? (next targets)) (not (visited (first targets))))
      nil
      (str "Go to " 
           (apply str (interpose " or " (map element-id targets)))
           " according to entrance conditions.")))) 

(defn create-activity-goto [activity visited]
  (let [target (first (map connection-target (transitions activity :out)))]
    (if (visited target)
      (str "Go to " (element-id target))
      nil)))

(defn condition-transition [trans]
  (let [const (first (mselect trans :constraint))]
    (if (nil? const)
      nil
      (first (mselect const :condition)))))

(defn condition-communication [comm]
  (let [flows (map representant-id (dataflows comm :input))
        part (representant-id (.eContainer (connection-source comm)))]
    (str "If \""
         (apply str (interpose "\" and \"" flows))
         "\" recieved from \"" part "\":")))
  
(defn create-activity-conditions [activity]
  (let [trans (transitions activity :in)
        ctrans (map #(str "If " %) (remove nil? (map condition-transition trans)))
        cflows (map condition-communication (communications activity :in))]
    (concat ctrans cflows)))

(defn report-communication-in [comm]
  (let [flows-in (map representant-id (dataflows comm :input))
        flows-out (map representant-id (dataflows comm :output))
        source-parent (representant-id (.eContainer (connection-source comm)))]
    (if (empty? flows-out)
      nil
      (str "Send \"" 
           (apply str (interpose "\", \"" flows-out)) 
           "\" to \"" 
           source-parent 
           "\" as response to \"" 
           (apply str (interpose "\", \"" flows-in))
           "\"."))))

(defn report-communication-out [comm]
  (let [flows-in (map representant-id (dataflows comm :output))
        flows-out (map representant-id (dataflows comm :input))
        target-parent (representant-id (.eContainer (connection-target comm)))]
    (cond (empty? flows-out)
          (str "Receive \"" 
               (apply str (interpose "\", \"" flows-in)) 
               "\" from \"" 
               target-parent 
               "\".")
          (empty? flows-in)
          (str "Send \"" 
               (apply str (interpose "\", \"" flows-out)) 
               "\" to \"" 
               target-parent 
               "\".")
          :else
          (str "Send \"" 
               (apply str (interpose "\", \"" flows-out)) 
               "\" to \"" 
               target-parent
               "\" and receive \""
               (apply str (interpose "\", \"" flows-in))
               "\" in response."))))

(defmulti node-report (fn [node visited] (type node)))

(defmethod node-report org.opencase.ordiagram.impl.StateImpl [state visited]
  (if (visited state)
    (Slot. [])
    (State. (representant-id state) (element-id state) nil (create-state-goto state visited))))

(defmethod node-report org.opencase.ordiagram.impl.ActivityImpl [activity visited]
  (if (visited activity)
    (Slot. [])
    (Activity. (representant-id activity) 
               (element-id activity) 
               nil 
               (create-activity-goto activity visited)
               (create-activity-conditions activity)
               (remove nil? (concat (map report-communication-in (communications activity :in))
                                    (map report-communication-out (communications activity :out)))))))

(defn service-activity-report [activity]
  (ServiceActivity. (representant-id activity)  
                    (create-activity-conditions activity)
                    (remove nil? (concat (map report-communication-in (communications activity :in))
                                         (map report-communication-out (communications activity :out))))))

(defn merge-levels [& levels]
  (if (nil? (next levels))
    (first levels)
    (let [cells (map :cell levels)]
      (Level. nil (Slot. cells)))))

(defn map-longest [fn missing-value & colls]
  (lazy-seq
    (when (not-every? empty? colls)
      (let [firsts (map first colls)]
        (cons
          (apply fn (map #(if (nil? %) missing-value %) firsts))
          (apply map-longest
            (conj (map rest colls) missing-value fn)))))))

(declare merge-trees)

(defn levels-from-successors [node visited]
  (if (visited node)
    ['() visited]
    (reduce merge-trees 
            ['() (conj visited node)] 
            (node-successors node))))

(defn merge-trees [[levels visited] node]  
  (let [[sub-levels new-visited] (levels-from-successors node visited)  
        new-levels (conj sub-levels (Level. nil (node-report node visited)))
        merged-levels (if (empty? levels)
                        new-levels
                        (doall (map-longest merge-levels (Level. nil (Slot. [])) levels new-levels)))]
    [merged-levels new-visited]))
            

(defmulti node-positions (fn [item lvl-index start-pos pos-map] (type item)))

(defmethod node-positions org.opencase.ordiagram.reporting.report.Level [level lvl-index start-pos pos-map]
  (let [[cell _ pos-map] (node-positions (:cell level) lvl-index start-pos pos-map)]
    [(Level. lvl-index cell) pos-map])) 

(defn cell-positions [[cells start-pos pos-map] lvl-index cell]
  (let [[cell new-start-pos new-pos-map] (node-positions cell lvl-index start-pos pos-map)]
    [(conj cells cell) new-start-pos new-pos-map])) 

(defmethod node-positions org.opencase.ordiagram.reporting.report.Slot [slot lvl-index start-pos pos-map]
  (let [[cells new-start-pos new-pos-map] (reduce #(cell-positions %1 lvl-index %2) [[] start-pos pos-map] (:cells slot))]
    [(assoc slot :cells cells) new-start-pos new-pos-map]))

(defmethod node-positions org.opencase.ordiagram.reporting.report.State [node lvl-index start-pos pos-map]
  [(assoc node :position (char start-pos)) (inc start-pos) (assoc pos-map (:id node) (str lvl-index (char start-pos)))])

(defmethod node-positions org.opencase.ordiagram.reporting.report.Activity [node lvl-index start-pos pos-map]
  [(assoc node :position (char start-pos)) (inc start-pos) (assoc pos-map (:id node) (str lvl-index (char start-pos)))])

(defn translate-goto-string [gstr pos-map]
  (if (nil? gstr)
    nil
    (clojure.string/replace gstr #"[0-9]+" #(str "§" (pos-map (Integer/parseInt %))))))

(defmulti node-links (fn [item pos] (type item)))

(defmethod node-links org.opencase.ordiagram.reporting.report.Level [level pos-map]
  (assoc level :cell (node-links (:cell level) pos-map)))

(defmethod node-links org.opencase.ordiagram.reporting.report.Slot [slot pos-map]
  (assoc slot :cells (vec (map #(node-links % pos-map) (:cells slot)))))

(defmethod node-links org.opencase.ordiagram.reporting.report.Activity [node pos-map]
  (assoc node :goto (translate-goto-string (:goto node) pos-map))) 

(defmethod node-links org.opencase.ordiagram.reporting.report.State [node pos-map]
  (assoc node :goto (translate-goto-string (:goto node) pos-map)))

(defn capitalize [text]
  (str (.toUpperCase (.substring text 0 1))
       (.toLowerCase (.substring text 1))))

(defn stateful-participant-report [participant]
  (let [[levels visited id-map] (merge-trees ['() #{} {}] (initial-node participant))
        lvl-col (map node-positions levels (map inc (range)) (repeat (int \a)) (repeat {}))
        [indexed pos-map] (reduce (fn [[indexed pos-map] [level new-map]] 
                                     (vector (conj indexed level) (merge pos-map new-map)))
                                  [[] {}] lvl-col)]
    (StatefulParticipant. (representant-id participant)
                          (capitalize (name (participant-type participant)))
                          (map #(node-links % pos-map) indexed))))
               
(defn service-participant-report [participant]
  (ServiceParticipant. (representant-id participant)
                       (capitalize (name (participant-type participant)))
                       (map service-activity-report (activities participant))))

(defn participant-report [participant]
  (let [nature (participant-nature participant)]
    (if (= nature :STATEFUL)
      (stateful-participant-report participant)
      (service-participant-report participant))))

(defn generate [or-diagram params] 
  (let [title        (representant-id or-diagram)
        parts        (participants or-diagram)
        part-reports (map participant-report parts)]
        ;msg          (create-or-message (count parts))]
    (ORDiagram. title part-reports))) 