(ns org.opencase.reporting.convertors.html
  (:use clojure.data.xml
        org.opencase.reporting.extensions))

(defmulti convert type) 

(defn funcs-from-element [element] 
  (let [ns   (.getAttribute element "namespace")
        convert-func (.getAttribute element "convert_func")
        css-func (.getAttribute element "css_func")]
    (require (symbol ns))
    {:convert (ns-resolve *ns* (symbol (str ns "/" convert-func)))
     :css     (ns-resolve *ns* (symbol (str ns "/" css-func)))}))

(defn add-element-to-map [map element]
  (let [format-key (keyword (.getAttribute element "format"))
        class-key  (.getAttribute element "class")]
    (merge-with merge 
                map 
                {format-key {class-key (funcs-from-element element)}})))

(defn create-convertors-map []
  (let [elements (configuration-elements-for "org.opencase.reporting.report_convertors")]
    (reduce add-element-to-map {} elements)))


(defn element-list [tag attr & colls]
  (apply element (list* tag attr (apply concat colls))))

(defn convert-collection [tag attrs col]
  (element-list tag attrs (map convert col)))

(defn convert-diagram [convertors-map diagram]
  (let [class-name (.getName (class diagram))
        convertor (get-in convertors-map [:HTML class-name :convert])]             
    (convertor diagram)))
           
(defn convert-diagrams [convertors-map diagrams]
  (element-list :div 
                {:class "diagrams"}
                (map convert-diagram (repeat convertors-map) diagrams)))

(defn load-resource [path]
  (let [class-loader (.getClassLoader (.getClass load-resource))]
    (clojure.java.io/resource path class-loader)))

(defn css-urls [convertors-map]
  (conj (map (fn [[class {css-loader :css}]] 
               (css-loader))
             (:HTML convertors-map))
        (load-resource "xhtml/report.css")))

(defn url-filename [url]
  (.getName (java.io.File. (.getFile url))))  

(defn stylesheet-links [filenames]
  (map #(element :link {:rel "stylesheet" :href %}) filenames))

(defn convert-report [report convertors-map stylesheet-urls]
  (let [css-filenames (map url-filename stylesheet-urls)]
    (element :html {:xmlns "http://www.w3.org/1999/xhtml"}
      (element-list :head {}
        [(element :meta {:http-equiv "content-type" :content "application/xhtml+xml; charset=utf-8"})]
         (stylesheet-links css-filenames)
        [(element :title {} (:title report))])
      (element :body {} 
        (element :div {:class "report-title"} (:title report))
        (convert-collection :div {:class "preamble"} (:preamble report))
        (convert-diagrams convertors-map (:diagrams report))))))
                             
(defmethod convert org.opencase.reporting.report.Paragraph [paragraph]
  (element :p {} (:text paragraph)))


(defn output-stylesheets [html-file urls]
  (let [css-paths (map #(str (.getParentFile html-file) java.io.File/separator (url-filename %)) urls)]
    (doseq [[path url] (map list css-paths urls)]
      (with-open [output-stream (java.io.FileOutputStream. path)
                  input-stream  (.getInputStream (.openConnection url))]
        (clojure.java.io/copy input-stream output-stream)))))

(defn create-output-file [dir] 
  (java.io.File. (str (.getAbsolutePath dir) java.io.File/separator "output.html"))) 

(defn produce [report params]
  (let [convertors-map (create-convertors-map)
        file (create-output-file (:output-dir params))
        stylesheet-urls (css-urls convertors-map)]    
    (with-open [writer (java.io.OutputStreamWriter. (java.io.FileOutputStream. file) "UTF-8")]
      (indent (convert-report report convertors-map stylesheet-urls) writer))
      ;(emit (transform report) writer :encoding "UTF-8"))
    (output-stylesheets file stylesheet-urls)))

