(ns org.opencase.clojure.modelapi
  (:use org.opencase.clojure.emfquery))

(defn element-id [element]
  "Returns unique instance ID of given element."
  (.getElementID element))

(defn representant-id [rep]
  "Returns ID of representants entity."
  (first (mselect rep :entity :id)))

(defn diagrams [project type] 
  "Returns diagrams from project of a specified type (e.g. :ORDiagram)."
  (let [type-filter (fn [col] 
                      (filter #(= (eclass %) type) col))]
    (mselect project :diagramCategories [:diagrams type-filter]))) 

(defn participants [or-diagram]
  "Returns participants of the OR diagram."  
  (values-of-class or-diagram :blocks :Participant))

(defn states [participant]
  "Returns states of participant."  
  (seq (.getStates participant)))

(defn activities [participant] 
  "Returns activities of participant."
  (seq (.getActivities participant)))

(defn notes [container] 
  "Returns notes of a container."
  (mselect container :notes))

(defn state-type [state]
  "Returns type of state for the given state: :start :INNER :FINAL :START_FINAL"
  (enum-to-keyword (.getType state)))

(defn participant-type [part]
  "Returns type of participant for the given participant: :PERSON :SYSTEM :ORGANISATION"
  (enum-to-keyword (.getType part)))

(defn participant-nature [part]
  "Returns nature of participant for the given participant: :STATEFUL :SERVICE"
  (enum-to-keyword (.getNature part)))

(defn connections [block direction]
  "Returns list of :in, :out or :all connections of given block."
  (case direction
    :in  (seq (.getConnectionsIn block))
    :out (seq (.getConnectionsOut block))
    :all (concat (connections block :in) 
                 (connections block :out))))

(defn connection-source [connection]
  "Returns source of the given connection."
  (.getSource connection))

(defn connection-target [connection]
  "Returns the target of the given connection."
  (.getTarget connection))

(defn communications [activity direction]
  "Returns a list of :in, :out or :all communications of the given activity."
  (filter #(= (eclass %) :Communication) 
          (connections activity direction)))

(defn transitions [node direction]
  "Returns a list of :in, :out or :all transitions of the given process node."
  (filter #(= (eclass %) :ORTransition) 
          (connections node direction)))

(defn start-state [process-holder]
  "Returns the start state of the given stateful process holder."
  (let [states (.getStartStates process-holder)] 
    (if (.isEmpty states)
      nil
      (.get states 0))))

(defn initial-activity [process-holder]
  "Returns the initial activity of the given stateful process holder."
  (let [activities (.getInitialActivities process-holder)] 
    (if (.isEmpty activities)
      nil
      (.get activities 0))))

(defn initial-node [process-holder]
  "Returns the initial node (state or node where the process begins) of the given stateful process holder."
  (let [node (start-state process-holder)]
    (if (nil? node)
      (initial-activity process-holder)
      node)))

(defn node-successors [node]
  "Returns a list of successors through transitions of the given process node."
  (map connection-target (transitions node :out)))

(defn dataflows [communication type]
  "Returns a list of :input, :output or :all data flows of the given communication."
  (let [dfs (mselect communication :labels)]
    (case type
      :input  (filter #(.isInput %) dfs)
      :output (remove #(.isInput %) dfs)
      :all    dfs)))

(defn activity-dataflows [activity direction]
  "Returns a list of :in, :out or :all data flows of the given activity."
  (case direction
    :in  (concat (mapcat #(dataflows % :input) (communications activity :in))
                 (mapcat #(dataflows % :output) (communications activity :out)))
    :out (concat (mapcat #(dataflows % :output) (communications activity :in))
                 (mapcat #(dataflows % :input) (communications activity :out)))
    :all (mapcat #(dataflows % :all) (communications activity :all))))

(defn constraint [connection]
  "Returns the constraint (or nil) of the given connection."
  (first (mselect connection :constraint)))
